@@ -20,6 +20,12 @@ from typing import Any, Optional
 import aiohttp
 
 from .config import API_BASE_URL, REFRESH_ENDPOINT, SIGN_IN_ENDPOINT
+from .exceptions import (
+    AuthenticationError,
+    InvalidCredentialsError,
+    TokenExpiredError,
+    TokenRefreshError,
+)
 
 __author__ = "Emmanuel Levijarvi"
 __copyright__ = "Emmanuel Levijarvi"
@@ -91,18 +97,94 @@ class AuthTokens:
 
     @classmethod
     def from_dict(cls, data: dict[str, Any]) -> "AuthTokens":
-        """Create AuthTokens from API response dictionary."""
+        """Create AuthTokens from API response dictionary or stored data.
+
+        Args:
+            data: Dictionary containing token data. Can be from API response
+                 (using camelCase keys) or from stored data (using snake_case
+                 keys from to_dict()).
+
+        Returns:
+            AuthTokens instance
+
+        Example:
+            # From API response
+            >>> tokens = AuthTokens.from_dict({
+            ...     "idToken": "...",
+            ...     "accessToken": "...",
+            ...     "refreshToken": "...",
+            ...     "authenticationExpiresIn": 3600
+            ... })
+
+            # From stored data (after to_dict())
+            >>> stored = tokens.to_dict()
+            >>> restored = AuthTokens.from_dict(stored)
+        """
+
+        # Helper to get value from either camelCase or snake_case key
+        def get_value(
+            camel_key: str, snake_key: str, default: Any = None
+        ) -> Any:
+            """Get value, checking camelCase first, then snake_case."""
+            value = data.get(camel_key)
+            if value is not None and value != "":
+                return value
+            value = data.get(snake_key)
+            if value is not None and value != "":
+                return value
+            return default
+
+        # Support both camelCase (API) and snake_case (stored) keys
         return cls(
-            id_token=data.get("idToken", ""),
-            access_token=data.get("accessToken", ""),
-            refresh_token=data.get("refreshToken", ""),
-            authentication_expires_in=data.get("authenticationExpiresIn", 3600),
-            access_key_id=data.get("accessKeyId"),
-            secret_key=data.get("secretKey"),
-            session_token=data.get("sessionToken"),
-            authorization_expires_in=data.get("authorizationExpiresIn"),
+            id_token=get_value("idToken", "id_token", ""),
+            access_token=get_value("accessToken", "access_token", ""),
+            refresh_token=get_value("refreshToken", "refresh_token", ""),
+            authentication_expires_in=get_value(
+                "authenticationExpiresIn", "authentication_expires_in", 3600
+            ),
+            access_key_id=get_value("accessKeyId", "access_key_id"),
+            secret_key=get_value("secretKey", "secret_key"),
+            session_token=get_value("sessionToken", "session_token"),
+            authorization_expires_in=get_value(
+                "authorizationExpiresIn", "authorization_expires_in"
+            ),
+            issued_at=datetime.fromisoformat(data["issued_at"])
+            if "issued_at" in data
+            else datetime.now(),
         )
 
+    def to_dict(self) -> dict[str, Any]:
+        """Convert AuthTokens to a dictionary for storage.
+
+        Returns a dictionary with all token data including the issued_at
+        timestamp, which is essential for correctly calculating expiration
+        times when restoring tokens.
+
+        Returns:
+            Dictionary with snake_case keys suitable for JSON serialization
+
+        Example:
+            >>> tokens = auth_client.current_tokens
+            >>> stored_data = tokens.to_dict()
+            >>> # Save to file/database
+            >>> import json
+            >>> json.dump(stored_data, file)
+            >>>
+            >>> # Later, restore tokens
+            >>> restored_tokens = AuthTokens.from_dict(json.load(file))
+        """
+        return {
+            "id_token": self.id_token,
+            "access_token": self.access_token,
+            "refresh_token": self.refresh_token,
+            "authentication_expires_in": self.authentication_expires_in,
+            "access_key_id": self.access_key_id,
+            "secret_key": self.secret_key,
+            "session_token": self.session_token,
+            "authorization_expires_in": self.authorization_expires_in,
+            "issued_at": self.issued_at.isoformat(),
+        }
+
     @property
     def expires_at(self) -> datetime:
         """Get the cached expiration timestamp."""
@@ -179,50 +261,20 @@ class AuthenticationResponse:
         )
 
 
-class AuthenticationError(Exception):
-    """Base exception for authentication errors.
-
-    Attributes:
-        message: Error message describing the failure
-        status_code: HTTP status code
-        response: Complete API response dictionary
-    """
-
-    def __init__(
-        self,
-        message: str,
-        status_code: Optional[int] = None,
-        response: Optional[dict[str, Any]] = None,
-    ):
-        """Initialize authentication error.
-
-        Args:
-            message: Error message describing the failure
-            status_code: HTTP status code
-            response: Complete API response dictionary
-        """
-        self.message = message
-        self.status_code = status_code
-        self.response = response
-        super().__init__(self.message)
-
-
-class InvalidCredentialsError(AuthenticationError):
-    """Raised when credentials are invalid."""
-
-    pass
-
-
-class TokenExpiredError(AuthenticationError):
-    """Raised when a token has expired."""
-
-    pass
-
-
-class TokenRefreshError(AuthenticationError):
-    """Raised when token refresh fails."""
-
-    pass
+# Exception classes moved to exceptions.py module
+# Import them here for backward compatibility
+__all__ = [
+    "UserInfo",
+    "AuthTokens",
+    "AuthenticationResponse",
+    "AuthenticationError",
+    "InvalidCredentialsError",
+    "TokenExpiredError",
+    "TokenRefreshError",
+    "NavienAuthClient",
+    "authenticate",
+    "refresh_access_token",
+]
 
 
 class NavienAuthClient:
@@ -236,7 +288,7 @@ class NavienAuthClient:
     - AWS credentials (if provided by API)
 
     Authentication is performed automatically when entering the async context
-    manager.
+    manager, unless valid stored tokens are provided.
 
     Example:
         >>> async with NavienAuthClient(user_id="user@example.com",
@@ -251,6 +303,16 @@ class NavienAuthClient:
         ...     if client.current_tokens.is_expired:
         ...         new_tokens = await
         client.refresh_token(client.current_tokens.refresh_token)
+
+        Restore session from stored tokens:
+        >>> stored_tokens = AuthTokens.from_dict(saved_data)
+        >>> async with NavienAuthClient(
+        ...     user_id="user@example.com",
+        ...     password="password",
+        ...     stored_tokens=stored_tokens
+        ... ) as client:
+        ...     # Authentication skipped if tokens are still valid
+        ...     print(f"Access token: {client.current_tokens.access_token}")
     """
 
     def __init__(
@@ -260,6 +322,7 @@ class NavienAuthClient:
         base_url: str = API_BASE_URL,
         session: Optional[aiohttp.ClientSession] = None,
         timeout: int = 30,
+        stored_tokens: Optional[AuthTokens] = None,
     ):
         """
         Initialize the authentication client.
@@ -270,10 +333,13 @@ class NavienAuthClient:
             base_url: Base URL for the API (default: official Navien API)
             session: Optional aiohttp ClientSession to use
             timeout: Request timeout in seconds
+            stored_tokens: Previously saved tokens to restore session.
+                          If provided and valid, skips initial sign_in.
 
         Note:
             Authentication is performed automatically when entering the
-            async context manager (using async with statement).
+            async context manager (using async with statement), unless
+            valid stored_tokens are provided.
         """
         self.base_url = base_url.rstrip("/")
         self._session = session
@@ -288,13 +354,44 @@ class NavienAuthClient:
         self._auth_response: Optional[AuthenticationResponse] = None
         self._user_email: Optional[str] = None
 
+        # Restore tokens if provided
+        if stored_tokens:
+            # Create a minimal AuthenticationResponse with stored tokens
+            # UserInfo will be populated on first API call if needed
+            self._auth_response = AuthenticationResponse(
+                user_info=UserInfo(
+                    user_type="",
+                    user_first_name="",
+                    user_last_name="",
+                    user_status="",
+                    user_seq=0,
+                ),
+                tokens=stored_tokens,
+            )
+            self._user_email = user_id
+
     async def __aenter__(self) -> "NavienAuthClient":
         """Async context manager entry."""
         if self._owned_session:
             self._session = aiohttp.ClientSession(timeout=self.timeout)
 
-        # Automatically authenticate
-        await self.sign_in(self._user_id, self._password)
+        # Check if we have valid stored tokens
+        if self._auth_response and self._auth_response.tokens:
+            tokens = self._auth_response.tokens
+            # If tokens are expired, refresh or re-authenticate
+            if tokens.are_aws_credentials_expired:
+                _logger.info(
+                    "Stored AWS credentials expired, re-authenticating..."
+                )
+                await self.sign_in(self._user_id, self._password)
+            elif tokens.is_expired:
+                _logger.info("Stored JWT token expired, refreshing...")
+                await self.refresh_token(tokens.refresh_token)
+            else:
+                _logger.info("Using stored tokens, skipping authentication")
+        else:
+            # No stored tokens, perform full authentication
+            await self.sign_in(self._user_id, self._password)
 
         return self
 
@@ -380,10 +477,12 @@ class NavienAuthClient:
 
         except aiohttp.ClientError as e:
             _logger.error(f"Network error during sign-in: {e}")
-            raise AuthenticationError(f"Network error: {str(e)}")
+            raise AuthenticationError(f"Network error: {str(e)}") from e
         except (KeyError, ValueError, json.JSONDecodeError) as e:
             _logger.error(f"Failed to parse authentication response: {e}")
-            raise AuthenticationError(f"Invalid response format: {str(e)}")
+            raise AuthenticationError(
+                f"Invalid response format: {str(e)}"
+            ) from e
 
     async def refresh_token(self, refresh_token: str) -> AuthTokens:
         """
@@ -466,10 +565,38 @@ class NavienAuthClient:
 
         except aiohttp.ClientError as e:
             _logger.error(f"Network error during token refresh: {e}")
-            raise TokenRefreshError(f"Network error: {str(e)}")
+            raise TokenRefreshError(f"Network error: {str(e)}") from e
         except (KeyError, ValueError, json.JSONDecodeError) as e:
             _logger.error(f"Failed to parse refresh response: {e}")
-            raise TokenRefreshError(f"Invalid response format: {str(e)}")
+            raise TokenRefreshError(f"Invalid response format: {str(e)}") from e
+
+    async def re_authenticate(self) -> AuthenticationResponse:
+        """
+        Re-authenticate using stored credentials.
+
+        This is a convenience method that uses the stored user_id and password
+        from initialization to perform a fresh sign-in. Useful for recovering
+        from expired tokens or connection issues.
+
+        Returns:
+            AuthenticationResponse with fresh tokens and user info
+
+        Raises:
+            ValueError: If stored credentials are not available
+            AuthenticationError: If authentication fails
+
+        Example:
+            >>> client = NavienAuthClient(email, password)
+            >>> await client.re_authenticate()  # Uses stored credentials
+        """
+        if not self.has_stored_credentials:
+            raise ValueError(
+                "No stored credentials available for re-authentication. "
+                "Credentials must be provided during initialization."
+            )
+
+        _logger.info("Re-authenticating with stored credentials")
+        return await self.sign_in(self._user_id, self._password)
 
     async def ensure_valid_token(self) -> Optional[AuthTokens]:
         """
@@ -526,6 +653,15 @@ class NavienAuthClient:
         """Get the email address of the authenticated user."""
         return self._user_email
 
+    @property
+    def has_stored_credentials(self) -> bool:
+        """Check if user credentials are stored for re-authentication.
+
+        Returns:
+            True if both user_id and password are available for re-auth
+        """
+        return bool(self._user_id and self._password)
+
     async def close(self) -> None:
         """Close the aiohttp session if we own it."""
         if self._owned_session and self._session:
