@@ -19,8 +19,18 @@ from typing import Any, Callable, Optional
 from awscrt import mqtt
 from awscrt.exceptions import AwsCrtError
 
-from .auth import NavienAuthClient
+from .auth import (
+    AuthenticationError,
+    NavienAuthClient,
+    TokenRefreshError,
+)
 from .events import EventEmitter
+from .exceptions import (
+    MqttConnectionError,
+    MqttCredentialsError,
+    MqttNotConnectedError,
+    MqttPublishError,
+)
 from .models import (
     Device,
     DeviceFeature,
@@ -133,17 +143,17 @@ class NavienMqttClient(EventEmitter):
                 credentials are not available
         """
         if not auth_client.is_authenticated:
-            raise ValueError(
+            raise MqttCredentialsError(
                 "Authentication client must be authenticated before "
                 "creating MQTT client. Call auth_client.sign_in() first."
             )
 
         if not auth_client.current_tokens:
-            raise ValueError("No tokens available from auth client")
+            raise MqttCredentialsError("No tokens available from auth client")
 
         auth_tokens = auth_client.current_tokens
         if not auth_tokens.access_key_id or not auth_tokens.secret_key:
-            raise ValueError(
+            raise MqttCredentialsError(
                 "AWS credentials not available in auth tokens. "
                 "Ensure authentication provides AWS IoT credentials."
             )
@@ -205,7 +215,8 @@ class NavienMqttClient(EventEmitter):
         # Schedule the coroutine in the stored loop using thread-safe method
         try:
             asyncio.run_coroutine_threadsafe(coro, self._loop)
-        except Exception as e:
+        except RuntimeError as e:
+            # Event loop is closed or not running
             _logger.error(f"Failed to schedule coroutine: {e}", exc_info=True)
 
     def _on_connection_interrupted_internal(
@@ -218,7 +229,6 @@ class NavienMqttClient(EventEmitter):
             error: Error that caused the interruption
             **kwargs: Forward-compatibility kwargs from AWS SDK
         """
-        _logger.warning(f"Connection interrupted: {error}")
         self._connected = False
 
         # Emit event
@@ -232,7 +242,7 @@ class NavienMqttClient(EventEmitter):
                 # Fallback for callbacks expecting no arguments
                 try:
                     self._on_connection_interrupted()  # type: ignore
-                except Exception as e:
+                except (TypeError, AttributeError) as e:
                     _logger.error(
                         f"Error in connection_interrupted callback: {e}"
                     )
@@ -339,12 +349,115 @@ class NavienMqttClient(EventEmitter):
                     "No connection manager available for reconnection"
                 )
 
-        except Exception as e:
+        except (AwsCrtError, AuthenticationError, RuntimeError) as e:
             _logger.error(
                 f"Error during active reconnection: {e}", exc_info=True
             )
             raise
 
+    async def _deep_reconnect(self) -> None:
+        """
+        Perform a deep reconnection by completely rebuilding the connection.
+
+        This method is called after multiple quick reconnection failures.
+        It performs a full teardown and rebuild:
+        - Disconnects existing connection
+        - Refreshes authentication tokens
+        - Creates new connection manager
+        - Re-establishes all subscriptions
+
+        This is more expensive but can recover from issues that a simple
+        reconnection cannot fix (e.g., stale credentials, corrupted state).
+        """
+        if self._connected:
+            _logger.debug("Already connected, skipping deep reconnection")
+            return
+
+        _logger.warning(
+            "Performing deep reconnection (full rebuild)... "
+            "This may take longer."
+        )
+
+        try:
+            # Step 1: Clean up existing connection if any
+            if self._connection_manager:
+                _logger.debug("Cleaning up old connection...")
+                try:
+                    if self._connection_manager.is_connected:
+                        await self._connection_manager.disconnect()
+                except (AwsCrtError, RuntimeError) as e:
+                    # Expected: connection already dead or in bad state
+                    _logger.debug(f"Error during cleanup: {e} (expected)")
+
+            # Step 2: Force token refresh to get fresh AWS credentials
+            _logger.debug("Refreshing authentication tokens...")
+            try:
+                # Use the stored refresh token from current tokens
+                current_tokens = self._auth_client.current_tokens
+                if current_tokens and current_tokens.refresh_token:
+                    await self._auth_client.refresh_token(
+                        current_tokens.refresh_token
+                    )
+                else:
+                    _logger.warning("No refresh token available")
+                    raise MqttCredentialsError(
+                        "No refresh token available for refresh"
+                    )
+            except (TokenRefreshError, ValueError, AuthenticationError) as e:
+                # If refresh fails, try full re-authentication with stored
+                # credentials
+                if self._auth_client.has_stored_credentials:
+                    _logger.warning(
+                        f"Token refresh failed: {e}. Attempting full "
+                        "re-authentication..."
+                    )
+                    await self._auth_client.re_authenticate()
+                else:
+                    _logger.error(
+                        "Cannot re-authenticate: no stored credentials"
+                    )
+                    raise
+
+            # Step 3: Create completely new connection manager
+            _logger.debug("Creating new connection manager...")
+            self._connection_manager = MqttConnection(
+                config=self.config,
+                auth_client=self._auth_client,
+                on_connection_interrupted=self._on_connection_interrupted_internal,
+                on_connection_resumed=self._on_connection_resumed_internal,
+            )
+
+            # Step 4: Attempt connection
+            success = await self._connection_manager.connect()
+
+            if success:
+                # Update connection references
+                self._connection = self._connection_manager.connection
+                self._connected = True
+
+                # Step 5: Re-establish subscriptions
+                if self._subscription_manager and self._connection:
+                    _logger.debug("Re-establishing subscriptions...")
+                    self._subscription_manager.update_connection(
+                        self._connection
+                    )
+                    await self._subscription_manager.resubscribe_all()
+
+                _logger.info(
+                    "Deep reconnection successful - fully rebuilt connection"
+                )
+            else:
+                _logger.error("Deep reconnection failed to connect")
+
+        except (
+            AwsCrtError,
+            AuthenticationError,
+            RuntimeError,
+            ValueError,
+        ) as e:
+            _logger.error(f"Error during deep reconnection: {e}", exc_info=True)
+            raise
+
     async def connect(self) -> bool:
         """
         Establish connection to AWS IoT Core.
@@ -394,6 +507,7 @@ class NavienMqttClient(EventEmitter):
                     is_connected_func=lambda: self._connected,
                     schedule_coroutine_func=self._schedule_coroutine,
                     reconnect_func=self._active_reconnect,
+                    deep_reconnect_func=self._deep_reconnect,
                     emit_event_func=self.emit,
                 )
                 self._reconnection_handler.enable()
@@ -428,7 +542,12 @@ class NavienMqttClient(EventEmitter):
 
             return False
 
-        except Exception as e:
+        except (
+            AwsCrtError,
+            AuthenticationError,
+            RuntimeError,
+            ValueError,
+        ) as e:
             _logger.error(f"Failed to connect: {e}")
             raise
 
@@ -439,7 +558,7 @@ class NavienMqttClient(EventEmitter):
         # Get current tokens from auth client
         auth_tokens = self._auth_client.current_tokens
         if not auth_tokens:
-            raise ValueError("No tokens available from auth client")
+            raise MqttCredentialsError("No tokens available from auth client")
 
         return AwsCredentialsProvider.new_static(
             access_key_id=auth_tokens.access_key_id,
@@ -473,7 +592,7 @@ class NavienMqttClient(EventEmitter):
             self._connection = None
 
             _logger.info("Disconnected successfully")
-        except Exception as e:
+        except (AwsCrtError, RuntimeError) as e:
             _logger.error(f"Error during disconnect: {e}")
             raise
 
@@ -493,7 +612,7 @@ class NavienMqttClient(EventEmitter):
 
         except json.JSONDecodeError as e:
             _logger.error(f"Failed to parse message payload: {e}")
-        except Exception as e:
+        except (AttributeError, KeyError, TypeError) as e:
             _logger.error(f"Error processing message: {e}")
 
     def _topic_matches_pattern(self, topic: str, pattern: str) -> bool:
@@ -555,7 +674,7 @@ class NavienMqttClient(EventEmitter):
             Exception: If subscription fails
         """
         if not self._connected or not self._subscription_manager:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         # Delegate to subscription manager
         return await self._subscription_manager.subscribe(topic, callback, qos)
@@ -574,7 +693,7 @@ class NavienMqttClient(EventEmitter):
             Exception: If unsubscribe fails
         """
         if not self._connected or not self._subscription_manager:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         # Delegate to subscription manager
         return await self._subscription_manager.unsubscribe(topic)
@@ -610,20 +729,19 @@ class NavienMqttClient(EventEmitter):
                 self._command_queue.enqueue(topic, payload, qos)
                 return 0  # Return 0 to indicate command was queued
             else:
-                raise RuntimeError("Not connected to MQTT broker")
+                raise MqttNotConnectedError("Not connected to MQTT broker")
 
         # Delegate to connection manager
         if not self._connection_manager:
-            raise RuntimeError("Connection manager not initialized")
+            raise MqttConnectionError("Connection manager not initialized")
 
         try:
             return await self._connection_manager.publish(topic, payload, qos)
-        except Exception as e:
+        except AwsCrtError as e:
             # Handle clean session cancellation gracefully
-            # Check exception type and name attribute for proper
-            # error identification
+            # Safely check e.name attribute (may not exist or be None)
             if (
-                isinstance(e, AwsCrtError)
+                hasattr(e, "name")
                 and e.name == "AWS_ERROR_MQTT_CANCELLED_FOR_CLEAN_SESSION"
             ):
                 _logger.warning(
@@ -638,12 +756,13 @@ class NavienMqttClient(EventEmitter):
                     self._command_queue.enqueue(topic, payload, qos)
                     return 0  # Return 0 to indicate command was queued
                 # Otherwise, raise an error so the caller can handle the failure
-                raise RuntimeError(
+                raise MqttPublishError(
                     "Publish cancelled due to clean session and "
-                    "command queue is disabled"
-                )
+                    "command queue is disabled",
+                    retriable=True,
+                ) from e
 
-            # Note: redact_topic is already used elsewhere in the file
+            # Other AWS CRT errors
             _logger.error(f"Failed to publish to topic: {e}")
             raise
 
@@ -663,7 +782,7 @@ class NavienMqttClient(EventEmitter):
             Subscription packet ID
         """
         if not self._connected or not self._subscription_manager:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         # Delegate to subscription manager
         return await self._subscription_manager.subscribe_device(
@@ -723,7 +842,7 @@ class NavienMqttClient(EventEmitter):
             ... )
         """
         if not self._connected or not self._subscription_manager:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         # Delegate to subscription manager (it handles state change
         # detection and events)
@@ -775,7 +894,7 @@ class NavienMqttClient(EventEmitter):
             ... )
         """
         if not self._connected or not self._subscription_manager:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         # Delegate to subscription manager
         return await self._subscription_manager.subscribe_device_feature(
@@ -793,7 +912,7 @@ class NavienMqttClient(EventEmitter):
             Publish packet ID
         """
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.request_device_status(device)
 
@@ -805,7 +924,7 @@ class NavienMqttClient(EventEmitter):
             Publish packet ID
         """
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.request_device_info(device)
 
@@ -823,7 +942,7 @@ class NavienMqttClient(EventEmitter):
             Publish packet ID
         """
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.set_power(device, power_on)
 
@@ -859,7 +978,7 @@ class NavienMqttClient(EventEmitter):
             - 5: Vacation Mode (requires vacation_days parameter)
         """
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.set_dhw_mode(
             device, mode_id, vacation_days
@@ -891,7 +1010,7 @@ class NavienMqttClient(EventEmitter):
             ValueError: If period_days is not in the valid range [1, 30]
         """
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.enable_anti_legionella(
             device, period_days
@@ -915,7 +1034,7 @@ class NavienMqttClient(EventEmitter):
             The message ID of the published command
         """
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.disable_anti_legionella(device)
 
@@ -946,7 +1065,7 @@ class NavienMqttClient(EventEmitter):
             await client.set_dhw_temperature(device, 120)
         """
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.set_dhw_temperature(
             device, temperature
@@ -988,7 +1107,7 @@ class NavienMqttClient(EventEmitter):
     ) -> int:
         """Update programmed reservations for temperature/mode changes."""
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.update_reservations(
             device, reservations, enabled=enabled
@@ -997,7 +1116,7 @@ class NavienMqttClient(EventEmitter):
     async def request_reservations(self, device: Device) -> int:
         """Request the current reservation program from the device."""
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.request_reservations(device)
 
@@ -1011,7 +1130,7 @@ class NavienMqttClient(EventEmitter):
     ) -> int:
         """Configure Time-of-Use pricing schedule via MQTT."""
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.configure_tou_schedule(
             device, controller_serial_number, periods, enabled=enabled
@@ -1024,7 +1143,7 @@ class NavienMqttClient(EventEmitter):
     ) -> int:
         """Request current Time-of-Use schedule from the device."""
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.request_tou_settings(
             device, controller_serial_number
@@ -1034,7 +1153,7 @@ class NavienMqttClient(EventEmitter):
         """Quickly toggle Time-of-Use functionality without
         modifying the schedule."""
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.set_tou_enabled(device, enabled)
 
@@ -1074,7 +1193,7 @@ class NavienMqttClient(EventEmitter):
             )
         """
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.request_energy_usage(
             device, year, months
@@ -1117,7 +1236,7 @@ class NavienMqttClient(EventEmitter):
             >>> await mqtt_client.request_energy_usage(device, 2025, [9])
         """
         if not self._connected or not self._subscription_manager:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         # Delegate to subscription manager
         return await self._subscription_manager.subscribe_energy_usage(
@@ -1135,7 +1254,7 @@ class NavienMqttClient(EventEmitter):
             Publish packet ID
         """
         if not self._connected or not self._device_controller:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         return await self._device_controller.signal_app_connection(device)
 
@@ -1180,7 +1299,9 @@ class NavienMqttClient(EventEmitter):
             - All tasks automatically stop when client disconnects
         """
         if not self._periodic_manager:
-            raise RuntimeError("Periodic request manager not initialized")
+            raise MqttConnectionError(
+                "Periodic request manager not initialized"
+            )
 
         await self._periodic_manager.start_periodic_requests(
             device, request_type, period_seconds
@@ -1210,7 +1331,9 @@ class NavienMqttClient(EventEmitter):
             >>> await mqtt_client.stop_periodic_requests(device)
         """
         if not self._periodic_manager:
-            raise RuntimeError("Periodic request manager not initialized")
+            raise MqttConnectionError(
+                "Periodic request manager not initialized"
+            )
 
         await self._periodic_manager.stop_periodic_requests(
             device, request_type
@@ -1242,7 +1365,9 @@ class NavienMqttClient(EventEmitter):
                 (default: 300 = 5 minutes)
         """
         if not self._periodic_manager:
-            raise RuntimeError("Periodic request manager not initialized")
+            raise MqttConnectionError(
+                "Periodic request manager not initialized"
+            )
 
         await self._periodic_manager.start_periodic_device_info_requests(
             device, period_seconds
@@ -1262,7 +1387,9 @@ class NavienMqttClient(EventEmitter):
                 (default: 300 = 5 minutes)
         """
         if not self._periodic_manager:
-            raise RuntimeError("Periodic request manager not initialized")
+            raise MqttConnectionError(
+                "Periodic request manager not initialized"
+            )
 
         await self._periodic_manager.start_periodic_device_status_requests(
             device, period_seconds
@@ -1278,7 +1405,9 @@ class NavienMqttClient(EventEmitter):
             device: Device object
         """
         if not self._periodic_manager:
-            raise RuntimeError("Periodic request manager not initialized")
+            raise MqttConnectionError(
+                "Periodic request manager not initialized"
+            )
 
         await self._periodic_manager.stop_periodic_device_info_requests(device)
 
@@ -1294,7 +1423,9 @@ class NavienMqttClient(EventEmitter):
             device: Device object
         """
         if not self._periodic_manager:
-            raise RuntimeError("Periodic request manager not initialized")
+            raise MqttConnectionError(
+                "Periodic request manager not initialized"
+            )
 
         await self._periodic_manager.stop_periodic_device_status_requests(
             device
@@ -1316,7 +1447,9 @@ class NavienMqttClient(EventEmitter):
             >>> await mqtt_client.stop_all_periodic_tasks()
         """
         if not self._periodic_manager:
-            raise RuntimeError("Periodic request manager not initialized")
+            raise MqttConnectionError(
+                "Periodic request manager not initialized"
+            )
 
         await self._periodic_manager.stop_all_periodic_tasks(_reason)
 
